/**
   Project line segments from view coordinates to clip coordinates.

   For each Model in a Scene, walk the Model's list of LineSegment objects.
   For each Vertex object in each LineSgement object, use the Scene's
   "projection matrix" to project the Vertex object into the front plane
   of the Scene's view volume.

   After projecting a vertex into the front plane of the view volume,
   "normalize" the coordinates of the vertex so that the front *face* of
   the Scene's view volume is represented by points with x and y coordinates
   between -1 and 1. (These new coordinates are called "clip coordinates" or
   "normalized device coordinates".)
*/

public class P2_View2Clip extends PipelineStage
{
   /**
      This constructor initializes this object's
      reference to the SceneRender object that
      contains it. This gives this stage access
      to addditional information that might be
      needed to process the Models passed to the
      processModel() method.
   */
   public P2_View2Clip(SceneRender sceneRender)
   {
      super(sceneRender);
   }


   /**
      For each line segment in the model, project each
      vertex to the front plane of the view volume and
      then transform each vertex from view coordinates
      to clip coordinates.
   */
   public Model processModel(Model model)
   {
      Camera camera = this.sceneRender.scene.camera;

      double left   = camera.left;   // These numbers define the camera's "view volume".
      double right  = camera.right;  // (In future renderers, these numbers will be
      double bottom = camera.bottom; //  encoded into a "projection matrix".)
      double top    = camera.top;
      double near   = camera.near;
      double far    = camera.far;

      // Project each line segment in the model's line segment list.
      for (LineSegment ls : model.lineSegmentList)
      {
         // For each vertex in the line segment, compute its projected (x,y) coordinates.
         double v0_xp = -near * (ls.v[0].x / ls.v[0].z);  // xp = near * (xv / -zv)
         double v0_yp = -near * (ls.v[0].y / ls.v[0].z);  // yp = near * (yv / -zv)
         double v1_xp = -near * (ls.v[1].x / ls.v[1].z);
         double v1_yp = -near * (ls.v[1].y / ls.v[1].z);

         if ( ! camera.perspective ) // do parallel projection instead
         {
            v0_xp = ls.v[0].x;
            v0_yp = ls.v[0].y;
            v1_xp = ls.v[1].x;
            v1_yp = ls.v[1].y;
         }

         // For each vertex, normalize the projected coordinates to "clip coordinates".
         double v0_xc = ((2 * v0_xp) - right - left)/(right - left);
         double v0_yc = ((2 * v0_yp) - top - bottom)/(top - bottom);
         double v1_xc = ((2 * v1_xp) - right - left)/(right - left);
         double v1_yc = ((2 * v1_yp) - top - bottom)/(top - bottom);

         // Mutate the Vertex objects so that they contain the (homogeneous) clip coordinates.
         ls.v[0].x = v0_xc;
         ls.v[0].y = v0_yc;
         ls.v[0].z = 0.0;
         ls.v[0].w = 1.0;
         ls.v[1].x = v1_xc;
         ls.v[1].y = v1_yc;
         ls.v[1].z = 0.0;
         ls.v[1].w = 1.0;
      }
      return model;
   }
}